import atexit
import sys
from time import sleep, time
from threading import Event
from math import atan2, degrees, hypot

if sys.version_info[0] > 2:
    from inspect import getfullargspec
else:
    from inspect import getargspec as getfullargspec

from .btcomm import BluetoothServer

class BlueDotPosition():
    """
    Represents a position of where the blue for is pressed, released or held.

    :param float x:
        The x position of the Blue Dot, 0 being centre, ``-1`` being far left 
        and ``1`` being far right. 

    :param float y:
        The y position of the Blue Dot, 0 being centre, ``-1`` being at the 
        bottom and ``1`` being at the top. 
    """
    def __init__(self, x, y):
        self._time = time()
        self._x = self._clamped(float(x))
        self._y = self._clamped(float(y))
        self._angle = None
        self._distance = None
        
    def _clamped(self, v):                                                                 
        return max(-1, min(1, v))
    
    @property
    def x(self):
        """
        The x position of the Blue Dot, ``0`` being centre, ``-1`` being far 
        left and ``1`` being far right. 
        """
        return self._x

    @property
    def y(self):
        """
        The y position of the Blue Dot, ``0`` being centre, ``-1`` being at 
        the bottom and ``1`` being at the top. 
        """
        return self._y

    @property
    def angle(self):
        """
        The angle from centre of where the Blue Dot is pressed, held or released.
        0 degress is up, 0 > 180 degrees clockwise, 0 > -180 degrees anti-clockwise.   
        """
        if self._angle == None:
            self._angle = degrees(atan2(self.x, self.y))
        return self._angle

    @property
    def distance(self):
        """
        The distance from centre of where the Blue Dot is pressed, held or released.
        The radius of the Blue Dot is ``1``.
        """
        if self._distance == None:
            self._distance = self._clamped(hypot(self.x, self.y))
        return self._distance

    @property
    def middle(self):
        """
        Returns ``True`` if the Blue Dot is pressed, held or released in the middle.
        """
        return True if self.distance <= 0.5 else False

    @property
    def top(self):
        """
        Returns ``True`` if the Blue Dot is pressed, held or released at the top.
        """
        return True if self.distance > 0.5 and self.angle > -45 and self.angle <= 45 else False
    
    @property
    def right(self):
        """
        Returns ``True`` if the Blue Dot is pressed, held or released on the right.
        """
        return True if self.distance > 0.5 and self.angle > 45 and self.angle <= 135 else False

    @property
    def bottom(self):
        """
        Returns ``True`` if the Blue Dot is pressed, held or released at the bottom.
        """        
        return True if self.distance > 0.5 and (self.angle > 135 or self.angle <= -135) else False

    @property
    def left(self):
        """
        Returns ``True`` if the Blue Dot is pressed, held or released on the left.
        """        
        return True if self.distance > 0.5 and self.angle > -135 and self.angle <= -45 else False

    @property
    def time(self):
        """
        The time the blue dot was at this position.

        Note - this is the time the message was received from the Blue Dot app, 
        not the time it was sent.
        """
        return self._time

class BlueDotInteraction():
    """
    Represents an interaction with the Blue Dot, from when it was pressed to 
    when it was released.

    A ``BlueDotInteraction`` can be active or inactive, i.e. it is active 
    because the Blue Dot has not been released, or inactive because the Blue 
    Dot was released and the interaction finished.

    :param BlueDotPosition pressed_position:
        The BlueDotPosition when the Blue Dot was pressed.
    """
    def __init__(self, pressed_position):
        self._active = False
        self._positions = []
        self._positions.append(pressed_position)

    @property
    def active(self):
        """
        Returns ``True`` if the interaction is still active, i.e. the Blue Dot 
        hasnt been released.
        """
        return self._active

    @property
    def positions(self):
        """
        A list of ``BlueDotPositions`` for all the positions which make up this
        interaction.

        The first position is where the Blue Dot was pressed, the last is where
        the Blue Dot was released, all position in between are where the position
        Blue Dot changed (i.e. moved) when it was held down.
        """
        return self._positions

    @property
    def pressed_position(self):
        """
        Returns the position when the Blue Dot was pressed i.e. where the 
        interaction started.
        """
        return self._positions[0]

    @property
    def released_position(self):
        """
        Returns the position when the Blue Dot was released i.e. where the 
        interaction ended.

        If the interaction is still active it returns ``None``.
        """
        if not self.active:
            return self._positions[-1]
        else:
            return None

    @property
    def current_position(self):
        """
        Returns the current position for the interaction.

        If the interaction is inactive, it will return the position when the
        Blue Dot was released
        """
        return self._positions[-1]

    @property
    def duration(self):
        """
        Returns the duration in seconds of the interaction, i.e. the amount time 
        between when the Blue Dot was pressed and now or when it was released.
        """
        if self.active:
            return time() - self.pressed_position.time
        else:
            return self.released_position.time - self.pressed_position.time

    @property
    def distance(self):
        """
        Returns the total distance of the Blue Dot interaction
        """
        dist = 0
        for i in range(1, len(self._positions)):
            p1 = self._positions[i-1]
            p2 = self._positions[i]
            dist += hypot(p2.x - p1.x, p2.y - p1.y)

        return dist

    def moved(self, moved_position):
        """
        Adds an additional position to the interaction, called when the position 
        the Blue Dot is pressed moves.
        """
        if self._active:
            self._positions.append(moved_position)
            
    def released(self, released_position):
        """
        Called when the Blue Dot is released and completes a Blue Dot interaction

        :param BlueDotPosition released_position:
            The BlueDotPosition when the Blue Dot was released.
        """
        self._active = False
        self._positions.append(released_position)

class BlueDotSwipe():
    """
    Represents a Blue Dot swipe interaction.

    A ``BlueDotSwipe`` can be valid or invalid based on whether the Blue Dot 
    interaction was a swipe or not.

    :param BlueDotInteraction interaction:
        The BlueDotInteraction object to be used to determine whether the interaction
        was a swipe.
    """
    def __init__(self, interaction):
        self._interaction = interaction
        self._speed_threshold = 2
        self._angle = None
        self._distance = None
        self._valid = self._is_valid_swipe()
    
    def _is_valid_swipe(self):
        #the validity of a swipe is based on the speed of the interaction, 
        # so a short fast swipe is valid as well as a long slow swipe 
        #self._speed = self.distance / self.interaction.duration
        self._speed = self.distance / self.interaction.duration
        if not self.interaction.active and self._speed > self._speed_threshold:
            return True
        else:
            return False

    @property
    def interaction(self):
        """
        The ``BlueDotInteraction`` object relating to this swipe.
        """
        return self._interaction
    
    @property
    def valid(self):
        """
        Returns ``True`` if the Blue Dot interaction is a swipe.
        """
        return self._valid

    @property
    def distance(self):
        """
        Returns the distance of the swipe (i.e. the distance between the pressed
        and released positions)
        """
        #should this be the total lenght of the swipe. All the points? It might be slow to calculate
        if self._distance == None:
            self._distance = hypot(
                self.interaction.released_position.x - self.interaction.pressed_position.x, 
                self.interaction.released_position.y - self.interaction.pressed_position.y)
        
        return self._distance

    @property
    def angle(self):
        """
        Returns the angle of the swipe (i.e. the angle between the pressed
        and released positions)
        """
        if self._angle == None:
            self._angle = degrees(atan2(
                self.interaction.released_position.x - self.interaction.pressed_position.x, 
                self.interaction.released_position.y - self.interaction.pressed_position.y))

        return self._angle

    @property
    def speed(self):
        """
        Returns the speed of the swipe in Blue Dot radius / second.
        """
        return self._speed
    
    @property
    def up(self):
        """
        Returns ``True`` if the Blue Dot was swiped up.
        """
        return True if self.valid and self.angle > -45 and self.angle <= 45 else False

    @property
    def down(self):
        """
        Returns ``True`` if the Blue Dot was swiped down.
        """        
        return True if self.valid and (self.angle > 135 or self.angle <= -135) else False

    @property
    def left(self):
        """
        Returns ``True`` if the Blue Dot was swiped left.
        """        
        return True if self.valid and self.angle > -135 and self.angle <= -45 else False

    @property
    def right(self):
        """
        Returns ``True`` if the Blue Dot was swiped right.
        """
        return True if self.valid and self.angle > 45 and self.angle <= 135 else False

class BlueDot():
    """
    Interacts with a Blue Dot client application, communicating when and where it
    has been pressed, released or held.   

    This class starts an instance of a bluetooth server (btcomm.BluetoothServer) 
    which manages the connection with the Blue Dot client.

    This class is intended for use with the Blue Dot client application.

    The following example will print a message when the Blue Dot is pressed::
    
        from bluedot import BlueDot
        bd = BlueDot()
        bd.wait_for_press()
        print("The blue dot was pressed")

    :param string device:
        The bluetooth device the server should use, the default is ``hci0``, if
        your device only has 1 bluetooth adapter this shouldn't need to be changed.

    :param int port:
        The bluetooth port the server should use, the default is ``1``, and under 
        normal use this should never need to change.

    :param bool auto_start_server:
        If ``True`` (the default), the bluetooth server will be automatically started
        on initialisation, if ``False``, the method ``start`` will need to use called
        before connections will be accepted.
    
    :param bool power_up_device:
        If ``True``, the bluetooth device will be powered up (if required) when the 
        server starts. The default is ``False``. 
        
        Depending on how bluetooth has been powered down, you may need to use rfkill 
        to unblock bluetooth to give permission to bluez to power on bluetooth::

            sudo rfkill unblock bluetooth

    :param bool print_messages:
        If ``True`` (the default), server status messages will be printed stating
        when the server has started and when clients connect / disconect.

    """
    def __init__(self, 
        device = "hci0", 
        port = 1,
        auto_start_server = True,
        power_up_device = False,
        print_messages = True):
        
        self._data_buffer = ""
        self._device = device
        self._port = port
        self._power_up_device = power_up_device
        self._print_messages = print_messages

        self._is_connected_event = Event()
        self._is_pressed_event = Event()
        self._is_released_event = Event()
        self._is_moved_event = Event()
        self._is_swiped_event = Event()
        self._is_double_pressed_event = Event()

        self._when_pressed = None
        self._when_double_pressed = None
        self._when_released = None
        self._when_moved = None
        self._when_swiped = None
        self._when_client_connects = None
        self._when_client_disconnects = None

        self._position = None
        self._interaction = None
        self._double_press_time = 0.3

        self._create_server()

        if auto_start_server:
            self.start()

    @property
    def device(self):
        """
        The bluetooth device the server is using. This defaults to ``hci0``.
        """
        return self._device
    
    @property
    def port(self):
        """
        The port the server is using. This defaults to ``1``.
        """
        return self._port

    @property
    def server(self):
        """
        The ``btcomm.BluetoothServer`` instance that is being used to communicate
        with clients.
        """
        return self._server

    @property
    def is_connected(self):
        """
        Returns ``True`` if a Blue Dot client is connected.
        """
        return self._is_connected_event.is_set()

    @property
    def is_pressed(self):
        """
        Returns ``True`` if the Blue Dot is pressed (or held).
        """
        return self._is_pressed_event.is_set()

    @property
    def value(self):
        """
        Returns a ``1`` if the Blue Dot is pressed, ``0`` if released.
        """
        return 1 if self.is_pressed else 0
        
    @property
    def values(self):
        """
        Returns an infinite generator constantly yielding the current value
        """
        while True:
            yield self.value

    @property
    def position(self):
        """
        Returns an instance of ``BlueDotPosition`` representing the 
        current or last position the Blue Dot was pressed, held or 
        released. 
        
        Note - if the Blue Dot is released (and inactive), ``position`` 
        will return position when it was released, until it is pressed 
        again. If the Blue Dot has never been pressed ``position`` will
        return ``None``.
        """
        return self._position

    @property
    def interaction(self):
        """
        Returns an instance of ``BlueDotInteraction`` representing the 
        current or last interaction with the Blue Dot. 
        
        Note - if the Blue Dot is released (and inactive), ``interaction`` 
        will return the interaction when it was released, until it is 
        pressed again. If the Blue Dot has never been pressed ``interaction`` 
        will return ``None``.
        """
        return self._interaction

    @property
    def when_pressed(self):
        """
        Sets or returns the function which is called when the Blue Dot is pressed. 

        The function should accept 0 or 1 parameters, if the function accepts 1 parameter an 
        instance of ``BlueDotPosition`` will be returned representing where the Blue Dot was pressed.
        
        The following example will print a message to the screen when the button is pressed::
        
            from bluedot import BlueDot
            
            def dot_was_pressed():
                print("The Blue Dot was pressed")
                
            bd = BlueDot()
            bd.when_pressed = dot_was_pressed
            
        This example shows how the position of where the dot was pressed can be obtained::
        
            from bluedot import BlueDot
            
            def dot_was_pressed(pos):
                print("The Blue Dot was pressed at pos x={} y={}".format(pos.x, pos.y))
                
            bd = BlueDot()
            bd.when_pressed = dot_was_pressed

        """
        return self._when_pressed

    @when_pressed.setter
    def when_pressed(self, value):
        self._when_pressed = value

    @property
    def when_double_pressed(self):
        """
        Sets or returns the function which is called when the Blue Dot is double pressed. 

        The function should accept 0 or 1 parameters, if the function accepts 1 parameter an 
        instance of ``BlueDotPosition`` will be returned representing where the Blue Dot was 
        pressed the second time.

        Note - the double press event is fired before the 2nd press event e.g. events would be 
        appear in the order, pressed, released, double pressed, pressed.
        """
        return self._when_double_pressed

    @when_double_pressed.setter
    def when_double_pressed(self, value):
        self._when_double_pressed = value

    @property 
    def double_press_time(self):
        """
        Sets or returns the time threshold in seconds for a double press. Defaults to ``0.3``.
        """
        return self._double_press_time

    @double_press_time.setter
    def double_press_time(self, value):
        self._double_press_time = value

    @property 
    def when_released(self):
        """
        Sets or returns the function which is called when the Blue Dot is released. 

        The function should accept 0 or 1 parameters, if the function accepts 1 parameter an 
        instance of ``BlueDotPosition`` will be returned representing where the Blue Dot was held 
        when it was released.
        """
        return self._when_released

    @when_released.setter
    def when_released(self, value):
        self._when_released = value

    @property
    def when_moved(self):
        """
        Sets or returns the function which is called when the position the Blue Dot is pressed is moved. 

        The function should accept 0 or 1 parameters, if the function accepts 1 parameter an 
        instance of ``BlueDotPosition`` will be returned representing the new position of where the 
        Blue Dot is held.
        """
        return self._when_moved

    @when_moved.setter
    def when_moved(self, value):
        self._when_moved = value

    @property
    def when_swiped(self):
        """
        Sets or returns the function which is called when the Blue Dot is swiped. 

        The function should accept 0 or 1 parameters, if the function accepts 1 parameter an 
        instance of ``BlueDotSwipe`` will be returned representing the how the Blue Dot was
        swiped.
        """
        return self._when_swiped

    @when_swiped.setter
    def when_swiped(self, value):
        self._when_swiped = value

    @property 
    def when_client_connects(self):
        """
        Sets or returns the function which is called when a Blue Dot connects. 
        """
        return self._when_client_connects

    @when_client_connects.setter
    def when_client_connects(self, value):
        self._when_client_connects = value

    @property 
    def when_client_disconnects(self):
        """
        Sets or returns the function which is called when a Blue Dot disconnects. 
        """
        return self._when_client_disconnects

    @when_client_disconnects.setter
    def when_client_disconnects(self, value):
        self._when_client_disconnects = value

    @property
    def print_messages(self):
        """
        When set to ``True`` results in messages relating to the status of the bluetooth server
        to be printed.
        """
        return self._print_messages

    @print_messages.setter
    def print_messages(self, value):
        self._print_messages = value

    def start(self):
        """
        Start the BluetoothServer if it is not already running. By default the server is started at
        initialisation.
        """            
        self._server.start()
        self._print_message("Server started {}".format(self.server.server_address))
        self._print_message("Waiting for connection")

    def _create_server(self):
        self._server = BluetoothServer(
                self._data_received, 
                when_client_connects = self._client_connected,
                when_client_disconnects = self._client_disconnected,
                device = self.device,
                port = self.port,
                power_up_device = self._power_up_device,
                auto_start = False)

    def stop(self):
        """
        Stop the bluetooth server.
        """
        self._server.stop()

    def wait_for_connection(self, timeout = None):
        """
        Waits until a Blue Dot client connects. 
        Returns ``True`` if a client connects. 

        :param float timeout:
            Number of seconds to wait for a wait connections, if ``None`` (the default), 
            it will wait indefinetly for a connection from a Blue Dot client.
        """
        return self._is_connected_event.wait(timeout)

    def wait_for_press(self, timeout = None):
        """
        Waits until a Blue Dot is pressed. 
        Returns ``True`` if the Blue Dot was pressed. 

        :param float timeout:
            Number of seconds to wait for a Blue Dot to be pressed, if ``None``
            (the default), it will wait indefinetly.
        """
        return self._is_pressed_event.wait(timeout)

    def wait_for_double_press(self, timeout = None):
        """
        Waits until a Blue Dot is double pressed. 
        Returns ``True`` if the Blue Dot was double pressed. 

        :param float timeout:
            Number of seconds to wait for a Blue Dot to be double pressed, if ``None``
            (the default), it will wait indefinetly.
        """
        return self._is_double_pressed_event.wait(timeout)

    def wait_for_release(self, timeout = None):
        """
        Waits until a Blue Dot is released. 
        Returns ``True`` if the Blue Dot was released. 

        :param float timeout:
            Number of seconds to wait for a Blue Dot to be released, if ``None`` 
            (the default), it will wait indefinetly.
        """
        return self._is_released_event.wait(timeout)

    def wait_for_move(self, timeout = None):
        """
        Waits until the position where the Blue Dot is pressed is moved. 
        Returns ``True`` if the position pressed on the Blue Dot was moved. 

        :param float timeout:
            Number of seconds to wait for the position that the Blue Dot 
            is pressed to move, if ``None`` (the default), it will wait indefinetly.
        """
        return self._is_moved_event.wait(timeout)

    def wait_for_swipe(self, timeout = None):
        """
        Waits until the Blue Dot is swiped. 
        Returns ``True`` if the Blue Dot was swiped. 

        :param float timeout:
            Number of seconds to wait for the Blue Dot to be swiped, if ``None``
            (the default), it will wait indefinetly.
        """
        return self._is_swiped_event.wait(timeout)

    def allow_pairing(self, timeout = 60):
        """
        Allow a Bluetooth device to pair with your Raspberry Pi by Putting the adapter 
        into discoverable and pairable mode.

        :param int timeout:
            The time in seconds the adapter will remain pairable. If set to ``None``
            the device will be discoverable and pairable indefinetly. 
        """
        self.server.adapter.allow_pairing(timeout = timeout)

    def _client_connected(self):
        self._is_connected_event.set()
        self._print_message("Client connected {}".format(self.server.client_address))
        if self.when_client_connects:
            self.when_client_connects()

    def _client_disconnected(self):
        self._is_connected_event.clear()
        self._print_message("Client disconnected")
        if self.when_client_disconnects:
            self.when_client_disconnects()
        
    def _data_received(self, data):
        #add the data received to the buffer
        self._data_buffer += data
        
        #get any full commands ended by \n
        last_command = self._data_buffer.rfind("\n")
        if last_command != -1:
            commands = self._data_buffer[:last_command].split("\n")
            self._process_commands(commands)
            #remove the processed commands from the buffer
            self._data_buffer = self._data_buffer[last_command + 1:]

    def _process_commands(self, commands):
        for command in commands:
            operation, x, y = command.split(",")
            self._position = BlueDotPosition(x, y)
        
            #dot released
            if operation == "0":
                self._released()
    
            #dot pressed
            elif operation == "1":
                self._pressed()

            #dot pressed position moved
            elif operation == "2":
                self._moved()

    def _pressed(self):
        self._is_pressed_event.set()
        self._is_released_event.clear()
        self._is_moved_event.clear()

        self._double_pressed()

        #create new interaction
        self._interaction = BlueDotInteraction(self._position)

        self._process_callback(self.when_pressed)

    def _double_pressed(self):
        #was there a previous interaction
        if self._interaction:
            #was it less than the time threshold (0.3 seconds)
            if self._interaction.duration < self._double_press_time:
                #was the dot pressed again in less than the threshold
                if time() - self._interaction.released_position.time < self._double_press_time:
                    self._is_double_pressed_event.set()

                    self._process_callback(self.when_double_pressed)

                    self._is_double_pressed_event.clear()

    def _released(self):
        self._is_pressed_event.clear()
        self._is_released_event.set()
        self._is_moved_event.clear()

        self._interaction.released(self._position)

        self._process_callback(self.when_released)

        self._process_interaction()

    def _moved(self):
        self._is_moved_event.set()

        self._interaction.moved(self._position)

        self._process_callback(self.when_moved)
    
        self._is_moved_event.clear()
    
    def _process_callback(self, callback):
        if callback:
            if len(getfullargspec(callback).args) == 0:
                callback()
            else:
                callback(self._position)

    def _process_interaction(self):
        #was the Blue Dot swiped?
        swipe = BlueDotSwipe(self._interaction)
        if swipe.valid:
            self._is_swiped_event.set()
            if self.when_swiped:
                if swipe.valid:
                    if len(getfullargspec(self.when_swiped).args) == 0:
                        self.when_swiped()
                    else:
                        self.when_swiped(swipe)

            self._is_swiped_event.clear()

    def _print_message(self, message):
        if self.print_messages:
            print(message)



from argparse import ArgumentParser
import pygame
import sys
from .btcomm import BluetoothAdapter, BluetoothClient

#colours
BLUE = (0, 0, 255)
DARKBLUE = (0, 0, 200)
GREY = (220, 220, 220)
RED = (255, 0, 0)

DEFAULTSIZE = (320, 240)
BORDER = 7
FONT = "monospace"
FONTSIZE = 18
FONTPAD = 3

class BlueDotClient():
    def __init__(self, device, server, fullscreen, width, height):

        #init pygame
        pygame.init()

        #load font
        font = pygame.font.SysFont(FONT, FONTSIZE)

        #setup the screen
        #set the screen caption
        pygame.display.set_caption("Blue Dot")

        #create the screen
        screenflags = 0
        
        if fullscreen:
            screenflags = pygame.FULLSCREEN
            if width == None and height == None:
                display_info = pygame.display.Info()
                width = display_info.current_w
                height = display_info.current_h

        if width == None: width = DEFAULTSIZE[0]
        if height == None: height = DEFAULTSIZE[1]
        
        screen = pygame.display.set_mode((width, height), screenflags)

        #has a server been specified?  If so connected directly
        if server:
            button_screen = ButtonScreen(screen, font, device, server, width, height)
            button_screen.run()
        else:
            #start the devices screen
            devices_screen = DevicesScreen(screen, font, device, width, height)
            devices_screen.run()

        pygame.quit()

class BlueDotScreen(object):
    def __init__(self, screen, font, width, height):
        self.screen = screen
        self.font = font
        self.width = width
        self.height = height

        #setup screen attributes
        self.frame_rect = pygame.Rect(BORDER, BORDER, self.width - (BORDER * 2), self.height - (BORDER * 2))
        self.close_rect = pygame.Rect(self.width - FONTSIZE - FONTPAD - BORDER, BORDER, FONTSIZE + FONTPAD, FONTSIZE + FONTPAD)

        self.draw_screen()
    
    def draw_screen(self):
        #set the screen background
        self.screen.fill(GREY)

        self.draw_close_button()
        
    def draw_close_button(self):
        #draw close button
        pygame.draw.rect(self.screen, BLUE, self.close_rect, 2)
        pygame.draw.line(self.screen, BLUE,
                        (self.close_rect[0], self.close_rect[1]),
                        (self.close_rect[0] + self.close_rect[2], self.close_rect[1] + self.close_rect[3]),
                        1)
        pygame.draw.line(self.screen, BLUE,
                        (self.close_rect[0], self.close_rect[1] + self.close_rect[3]),
                        (self.close_rect[0] + self.close_rect[2], self.close_rect[1]),
                        1)
        
    def draw_error(self, e):
        message = "Error: {}".format(e)
        print(message)
        self.draw_status_message(message, colour = RED)

    def draw_status_message(self, message, colour = BLUE):
        self.screen.fill(GREY, self.frame_rect)
        self.draw_close_button()
        self.draw_text(message, colour, self.frame_rect.height / 2, border = True, border_pad = FONTPAD)
        pygame.display.update()

    def draw_text(self, text, colour, start_y, antiaalias=False, background=None, border=False, border_width=1, border_pad=0):
        rect = pygame.Rect(self.frame_rect)
        y = rect.top + start_y + border_pad
        lineSpacing = -2

        # get the height of the font
        fontHeight = self.font.size("Tg")[1]

        while text:
            i = 1

            # determine if the row of text will be outside our area
            if y + fontHeight > rect.bottom:
                break

            # determine maximum width of line
            while self.font.size(text[:i])[0] < (rect.width - (border_pad * 2)) and i < len(text):
                i += 1

            # if we've wrapped the text, then adjust the wrap to the last word      
            if i < len(text): 
                i = text.rfind(" ", 0, i) + 1

            # render the line and blit it to the surface
            if background:
                image = self.font.render(text[:i], 1, colour, background)
                image.set_colorkey(background)
            else:
                image = self.font.render(text[:i], antiaalias, colour)

            self.screen.blit(image, (rect.left + border_pad, y))
            y += fontHeight + lineSpacing + border_pad

            # remove the text we just blitted
            text = text[i:]

        #return the rect the text was drawn in
        rect.top = rect.top + start_y
        rect.height = y - start_y
    
        if border:
            pygame.draw.rect(self.screen, colour, rect, border_width)

        return rect

class DevicesScreen(BlueDotScreen):
    def __init__(self, screen, font, device, width, height):
        self.bt_adapter = BluetoothAdapter(device = device)

        super(DevicesScreen, self).__init__(screen, font, width, height)
        #self.draw_screen()

    def draw_screen(self):
        self.device_rects = []

        super(DevicesScreen, self).draw_screen()

        #title
        title_rect = self.draw_text("Connect", RED, 0)

        y = title_rect.bottom
        for d in self.bt_adapter.paired_devices:
            device_rect = self.draw_text("{} ({})".format(d[1], d[0]), BLUE, y, border = True, border_pad = FONTPAD)
    
            self.device_rects.append(pygame.Rect(device_rect))

            y = device_rect.bottom

    def run(self):
        clock = pygame.time.Clock()
        running = True
        while running:
            clock.tick(50)

            ev = pygame.event.get()

            for event in ev:

                if event.type == pygame.MOUSEBUTTONDOWN:
                    pos = pygame.mouse.get_pos()
                    
                    #has a device been clicked?
                    for d in range(len(self.device_rects)):
                        if self.device_rects[d].collidepoint(pos):
                            # show the button
                            self.draw_status_message("Connecting")
                            button_screen = ButtonScreen(self.screen, self.font, self.bt_adapter.device, self.bt_adapter.paired_devices[d][0], self.width, self.height)
                            button_screen.run()

                            #redraw the screen
                            self.draw_screen()

                    if self.close_rect.collidepoint(pos):
                        running = False
                    
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        running = False

                if event.type == pygame.QUIT:
                    running = False
                    
            pygame.display.update()

class ButtonScreen(BlueDotScreen):
    def __init__(self, screen, font, device, server, width, height):
        super(ButtonScreen, self).__init__(screen, font, width, height)

        self.device = device
        self.server = server

    def draw_screen(self):
        super(ButtonScreen, self).draw_screen()

        self._draw_circle(BLUE)

    def _draw_circle(self, colour):
        #draw the circle
        self.circle_centre = (int(self.frame_rect.top + (self.frame_rect.width / 2)), int(self.frame_rect.left + (self.frame_rect.height / 2)))
        if self.frame_rect.width > self.frame_rect.height:
            self.circle_radius = int(self.frame_rect.height / 2)
        else:
            self.circle_radius = int(self.frame_rect.width / 2)
        
        self.circle_rect = pygame.draw.circle(self.screen, colour, self.circle_centre, self.circle_radius, 0)

    def _send_message(self, op, pos):
        if self.bt_client.connected:
            x = (pos[0] - self.circle_centre[0]) / float(self.circle_radius)
            y = ((pos[1] - self.circle_centre[1]) / float(self.circle_radius)) * -1
            message = "{},{},{}\n".format(op, x, y)
            try:
                self.bt_client.send(message)
            except:
                e = str(sys.exc_info()[1])
                self.draw_error(e)
        else:
            self.draw_error("Blue Dot not connected")
            
    def run(self):
        
        self.bt_client = BluetoothClient(self.server, None, device = self.device, auto_connect = False)
        try:
            self.bt_client.connect()
        except:
            e = str(sys.exc_info()[1])
            self.draw_error(e)
    
        clock = pygame.time.Clock()
        pygame.event.clear()
        
        mouse_pressed = False
        running = True
        while running:
            clock.tick(50)

            ev = pygame.event.get()

            for event in ev:
                
                # handle mouse
                if event.type == pygame.MOUSEBUTTONDOWN or event.type == pygame.MOUSEBUTTONUP or (event.type == pygame.MOUSEMOTION and mouse_pressed):
                    pos = pygame.mouse.get_pos()

                    #circle clicked?
                    if self.circle_rect.collidepoint(pos):

                        if event.type == pygame.MOUSEBUTTONDOWN:
                            mouse_pressed = True
                            self._draw_circle(DARKBLUE)
                            self._send_message(1, pos)
                            
                        elif event.type == pygame.MOUSEBUTTONUP:
                            mouse_pressed = False
                            self._draw_circle(BLUE)
                            self._send_message(0, pos)
                        
                        elif event.type == pygame.MOUSEMOTION:
                            self._send_message(2, pos)
                
                    #close clicked?
                    if self.close_rect.collidepoint(pos):
                        running = False
                    
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        running = False

                if event.type == pygame.QUIT:
                    running = False
                    
            pygame.display.update()
            
        self.bt_client.disconnect()

if __name__ == "__main__":
    #read command line options
    parser = ArgumentParser(description="Blue Dot Python App")
    parser.add_argument("--device", help="The name of the bluetooth device to use (default is hci0)")
    parser.add_argument("--server", help="The name or mac address of the bluedot server")
    parser.add_argument("--fullscreen", help="Fullscreen app", action="store_true")
    parser.add_argument("--width", type=int, help="A custom screen width (default is {})".format(DEFAULTSIZE[0]))
    parser.add_argument("--height", type=int, help="A customer screen height (default is {})".format(DEFAULTSIZE[1]))
    args = parser.parse_args()

    #start the blue dot client
    blue_dot_client = BlueDotClient(args.device, args.server, args.fullscreen, args.width, args.height)





from .btcomm import BluetoothServer
from .dot import BlueDot
from .threads import WrapThread

class MockBluetoothAdapter():
    def __init__(self, device = "mock0"):
        self._device = device
        self._address = "00:00:00:00:00:00"
        self._powered = True
        self._discoverable = False
        self._pairable = False
        self._pairing_thread = None

    @property
    def device(self):
        return self._device
    
    @property
    def address(self):
        return self._address

    @property
    def powered(self):
        return self._powered

    @powered.setter
    def powered(self, value):
        self._powered = value

    @property
    def discoverable(self):
        return self._discoverable

    @discoverable.setter
    def discoverable(self, value):
        self._discoverable = value

    @property
    def pairable(self):
        return self._pairable

    @pairable.setter
    def pairable(self, value):
        self._pairable = value

    @property
    def paired_devices(self):
        return [["01:01:01:01:01:01", "mock_device_1"], ["02:02:02:02:02:02", "mock_device_2"]]

    def allow_pairing(self, timeout = 60):
        #if a pairing thread is already running, stop it and restart
        if self._pairing_thread:
            if self._pairing_thread.is_alive:
                self._pairing_thread.stop()
        
        #start the pairing thread
        self._pairing_thread = WrapThread(target=self._allow_pairing, args=(timeout, ))
        self._pairing_thread.start()

    def _allow_pairing(self, timeout):
        self.pairable = True
        self.discoverable = True
        if timeout != None:
            #wait till the timeout or the thread is stopped
            self._pairing_thread.stopping.wait(timeout)
            self.discoverable = False
            self.pairable = False


class MockBluetoothServer(BluetoothServer):
    """
    MockBluetoothServer inherits from BluetoothServer but overrides __init__, .start 
    and .stop to create a MockBluetoothServer which can be used for testing and debugging.
    """
    def __init__(self, 
        data_received_callback, 
        auto_start = True, 
        device = "mock0", 
        port = 1,
        encoding = "utf-8",
        power_up_device = False,
        when_client_connects = None, 
        when_client_disconnects = None):
    
        self._device = device
        self._adapter = MockBluetoothAdapter(self._device)

        self._data_received_callback = data_received_callback
        self._port = port
        self._encoding = encoding
        self._power_up_device = power_up_device
        self._when_client_connects = when_client_connects
        self._when_client_disconnects = when_client_disconnects

        self._running = False
        self._client_connected = False
        self._server_sock = None
        self._client_info = None
        self._client_sock = None
            
    def start(self):
        self._running = True
        
    def stop(self):
        self._running = False
        
    def mock_client_connected(self, client_address = "11:11:11:11:11:11"):
        """
        Simulates a client connected to the BluetoothServer. 

        :param string client_address:
            The mock client mac address, defaults to '11:11:11:11:11:11'
        """
        if not self._client_connected:
            self._client_connected = True
            self._client_info = (client_address, self.port)
            #call the call back
            if self.when_client_connects:
                self._when_client_connects()
                
    def mock_client_disconnected(self):
        """
        Simulates a client disconnecting from the BluetoothServer. 
        """
        if self._client_connected:
            self._client_connected = False
            self._client_info = None
            if self._when_client_disconnects:
                self._when_client_disconnects()

    def mock_client_sending_data(self, data):
        """
        Simulates a client sending data to the BluetoothServer. 
        """
        if self._client_connected:
            self._data_received_callback(data)

class MockBlueDot(BlueDot):
    """
    MockBlueDot inherits from BlueDot but overrides ._create_server, to create
    a MockBluetoothServer which can be used for testing and debugging.
    """
    def _create_server(self):
        self._server = MockBluetoothServer(
                self._data_received, 
                when_client_connects = self._client_connected,
                when_client_disconnects = self._client_disconnected,
                device = self.device,
                port = self.port)

    def mock_client_connected(self, client_address = "11:11:11:11:11:11"):
        """
        Simulates a client connecting to the BlueDot. 

        :param string client_address:
            The mock client mac address, defaults to '11:11:11:11:11:11'
        """
        self._server.mock_client_connected(client_address)
                
    def mock_client_disconnected(self):
        """
        Simulates a client disconnecting from the BlueDot. 
        """
        self._server.mock_client_disconnected()
                
    def mock_blue_dot_pressed(self, x, y):
        """
        Simulates the blue dot being pressed. 

        :param int x:
            The x position where the mock blue dot was pressed

        :param int y:
            The y position where the mock blue dot was pressed    
        """
        self._server.mock_client_sending_data("1,{},{}\n".format(x, y))

    def mock_blue_dot_released(self, x, y):
        """
        Simulates the blue dot being released. 

        :param int x:
            The x position where the mock blue dot was released

        :param int y:
            The y position where the mock blue dot was released   
        """
        self._server.mock_client_sending_data("0,{},{}\n".format(x, y))

    def mock_blue_dot_moved(self, x, y):
        """
        Simulates the blue dot being moved. 

        :param int x:
            The x position where the mock blue dot was moved too

        :param int y:
            The y position where the mock blue dot was moved too  
        """
        self._server.mock_client_sending_data("2,{},{}\n".format(x, y))

    def launch_mock_app(self):
        """
        Launches a mock Blue Dot app.
        
        The mock app reacts to mouse clicks and movement and calls the mock blue
        dot methods to simulates presses.
        
        This is useful for testing, allowing you to interact with BlueDot without
        having to script mock functions.
        
        The mock app uses pygame which will need to be installed.
        """
        self._mock_app_thread = WrapThread(target=self._launch_mock_app)
        self._mock_app_thread.start()

    def _launch_mock_app(self):
        #imported here, so pygame is only a pre-requisite for the mock app
        import pygame
        
        pygame.init()
        screen = pygame.display.set_mode((200,200))
        pygame.display.set_caption("Blue Dot")
        
        clock = pygame.time.Clock()

        circle_centre = (100, 100)
        circle_radius = 100
        circle_rect = pygame.draw.circle(screen, (0,0, 255), circle_centre, circle_radius, 0)
        
        self.mock_client_connected()

        running = True
        while running:
            clock.tick(50)
        
            # get all events
            ev = pygame.event.get()

            # proceed events
            for event in ev:

                # handle mouse
                if event.type == pygame.MOUSEBUTTONDOWN or event.type == pygame.MOUSEBUTTONUP or (event.type == pygame.MOUSEMOTION and self.is_pressed):
                    pos = pygame.mouse.get_pos()
                    
                    if circle_rect.collidepoint(pos):
                        x = (pos[0] - circle_centre[0]) / circle_radius
                        y = ((pos[1] - circle_centre[1]) / circle_radius) * -1

                        if event.type == pygame.MOUSEBUTTONDOWN:
                            self.mock_blue_dot_pressed(x,y)

                        elif event.type == pygame.MOUSEBUTTONUP:
                            self.mock_blue_dot_released(x,y)
                        
                        elif event.type == pygame.MOUSEMOTION:
                            self.mock_blue_dot_moved(x,y)

                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        running = False

                if event.type == pygame.QUIT:
                    running = False

                if self._mock_app_thread.stopping.is_set():
                    running = False
                
            pygame.display.update()

        self.mock_client_disconnected()

        pygame.quit()


import socket
import sys
from time import sleep

from .utils import register_spp, get_mac, get_adapter_powered_status, get_adapter_discoverable_status, get_adapter_pairable_status, get_paired_devices, device_pairable, device_discoverable, device_powered, string_to_bytes

from .threads import WrapThread

if sys.version_info[0] > 2:
    BLUETOOTH_EXCEPTIONS = (BlockingIOError, ConnectionResetError, TimeoutError)
else:
    BLUETOOTH_EXCEPTIONS = (IOError)

BLUETOOTH_TIMEOUT = 0.01

class BluetoothAdapter():
    """
    Represents and allows interaction with a Bluetooth Adapter.

    The following example will get the bluetooth adapter, print its powered status
    and any paired devices::

        a = BluetoothAdapter()
        print("Powered = {}".format(a.powered))
        print(a.paired_devices)

    :param string device:
        The bluetooth device to be used, the default is ``hci0``, if your device 
        only has 1 bluetooth adapter this shouldn't need to be changed.
    """
    def __init__(self, device = "hci0"):
        self._device = device
        self._address = get_mac(self._device)
        self._pairing_thread = None

    @property
    def device(self):
        """
        The bluetooth device name. This defaults to hci0.
        """
        return self._device
    
    @property
    def address(self):
        """
        The mac address of the bluetooth adapter.
        """
        return self._address

    @property
    def powered(self):
        """
        Set to ``True`` to power on the bluetooth adapter.

        Depending on how bluetooth has been powered down, you may need to use 
        rfkill to unblock bluetooth to give permission to bluez to power on bluetooth::

            sudo rfkill unblock bluetooth
        """
        return get_adapter_powered_status(self._device)

    @powered.setter
    def powered(self, value):
        device_powered(self._device, value)

    @property
    def discoverable(self):
        """
        Set to ``True`` to make the bluetooth adapter discoverable.
        """
        return get_adapter_discoverable_status(self._device)

    @discoverable.setter
    def discoverable(self, value):
        device_discoverable(self._device, value)

    @property
    def pairable(self):
        """
        Set to ``True`` to make the bluetooth adapter pairable.
        """
        return get_adapter_pairable_status(self._device)

    @pairable.setter
    def pairable(self, value):
        device_pairable(self._device, value)

    @property
    def paired_devices(self):
        """
        Returns a list of devices paired with this adapater 
        ``((device_mac_address, device_name), (device_mac_address, device_name))``::
        
            a = BluetoothAdapter()
            devices = a.paired_devices
            for d in devices:
                device_address = d[0]
                device_name = d[1]

        """
        return get_paired_devices(self._device)

    def allow_pairing(self, timeout = 60):
        """
        Put the adapter into discoverable and pairable mode.

        :param int timeout:
            The time in seconds the adapter will remain pairable. If set to ``None``
            the device will be discoverable and pairable indefinetly. 
        """
        #if a pairing thread is already running, stop it and restart
        if self._pairing_thread:
            if self._pairing_thread.is_alive:
                self._pairing_thread.stop()
        
        #make the adapter pairable
        self.pairable = True
        self.discoverable = True
        
        if timeout != None:
            #start the pairing thread
            self._pairing_thread = WrapThread(target=self._expire_pairing, args=(timeout, ))
            self._pairing_thread.start()

    def _expire_pairing(self, timeout):
        #wait till the timeout or the thread is stopped
        self._pairing_thread.stopping.wait(timeout)
        self.discoverable = False
        self.pairable = False

class BluetoothServer():
    """
    Creates a Bluetooth server which will allow connections and accept incoming 
    RFCOMM serial data.

    When data is received by the server it is passed to a callback function
    which must be specified at initiation.

    The following example will create a bluetooth server which will wait for a 
    connection and print any data it receives and send it back to the client::
    
        from bluedot.btcomm import BluetoothServer
        from signal import pause

        def data_received(data):
            print(data)
            s.send(data)

        s = BluetoothServer(data_received)
        pause()

    :param data_received_callback:
        A function reference should be passed, this function will be called when
        data is received by the server.  The function should accept a single parameter
        which when called will hold the data received. Set to ``None`` if  received 
        data is not required.

    :param bool auto_start:
        If ``True`` (the default), the bluetooth server will be automatically started
        on initialisation, if ``False``, the method ``start`` will need to be called
        before connections will be accepted.

    :param string device:
        The bluetooth device the server should use, the default is ``hci0``, if
        your device only has 1 bluetooth adapter this shouldn't need to be changed.

    :param int port:
        The bluetooth port the server should use, the default is ``1``.

    :param string encoding:
        The encoding standard to be used when sending and receiving byte data. The default is 
        ``utf-8``.  If set to ``None`` no encoding is done and byte data types should be used.

    :param bool power_up_device:
        If ``True``, the bluetooth device will be powered up (if required) when the 
        server starts. The default is ``False``. 
        
        Depending on how bluetooth has been powered down, you may need to use rfkill 
        to unblock bluetooth to give permission to bluez to power on bluetooth::

            sudo rfkill unblock bluetooth

    :param when_client_connects:
        A function reference which will be called when a client connects. If ``None``
        (the default), no notification will be given when a client connects

    :param when_client_disconnects:
        A function reference which will be called when a client disconnects. If ``None``
        (the default), no notification will be given when a client disconnects

    """
    def __init__(self, 
        data_received_callback, 
        auto_start = True, 
        device = "hci0", 
        port = 1,
        encoding = "utf-8",
        power_up_device = False,
        when_client_connects = None, 
        when_client_disconnects = None):

        self._device = device
        self._adapter = BluetoothAdapter(self._device)

        self._data_received_callback = data_received_callback
        self._port = port
        self._encoding = encoding
        self._power_up_device = power_up_device
        self._when_client_connects = when_client_connects
        self._when_client_disconnects = when_client_disconnects

        self._running = False
        self._client_connected = False
        self._server_sock = None
        self._client_info = None
        self._client_sock = None
        
        self._conn_thread = None

        if auto_start:
            self.start()
    
    @property
    def device(self):
        """
        The bluetooth device the server is using. This defaults to hci0.
        """
        return self.adapter.device

    @property
    def adapter(self):
        """
        A BluetoothAdapter object which represents the bluetooth device 
        the server is using.
        """
        return self._adapter

    @property
    def port(self):
        """
        The port the server is using. This defaults to 1.
        """
        return self._port

    @property
    def encoding(self):
        """
        The encoding standard the server is using. This defaults to ``utf-8``.
        """
        return self.encoding

    @property
    def running(self):
        """
        Returns a ``True`` if the server is running.
        """
        return self._running

    @property
    def server_address(self):
        """
        The mac address of the device the server is using.
        """
        return self.adapter.address

    @property
    def client_address(self):
        """
        The mac address of the client connected to the server. Returns 
        ``None`` if no client is connected.
        """
        if self._client_info:
            return self._client_info[0]
        else:
            return None

    @property
    def client_connected(self):
        """
        Returns ``True`` if a client is connected.
        """
        return self._client_connected

    @property
    def data_received_callback(self):
        """
        Sets or returns the function which is called when data is received by the server. 

        The function should accept a single parameter which when called will hold
        the data received. Set to ``None`` if received data is not required.
        """
        return self._data_received_callback

    @data_received_callback.setter
    def data_received_callback(self, value):
        self._data_received_callback = value

    @property
    def when_client_connects(self):
        """
        Sets or returns the function which is called when a client connects. 
        """
        return self._when_client_connects

    @when_client_connects.setter
    def when_client_connects(self, value):
        self._when_client_connects = value

    @property
    def when_client_disconnects(self):
        """
        Sets or returns the function which is called when a client disconnects. 
        """
        return self._when_client_disconnects

    @when_client_disconnects.setter
    def when_client_disconnects(self, value):
        self._when_client_disconnects = value

    def start(self):
        """
        Starts the Bluetooth server if its not already running. The server needs to be started before
        connections can be made.
        """
        if not self._running:

            if self._power_up_device:
                self.adapter.powered = True

            if not self.adapter.powered:
                raise Exception("Bluetooth device {} is turned off".format(self.adapter.device))

            #register the serial port profile with bluetooth
            register_spp(self._port)

            #start bluetooth server
            #open the bluetooth socket
            self._server_sock = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_STREAM, socket.BTPROTO_RFCOMM)
            self._server_sock.settimeout(BLUETOOTH_TIMEOUT)
            try:
                self._server_sock.bind((self.server_address, self.port))
            except (socket.error, OSError) as e:
                if str(e) == "[Errno 98] Address already in use":
                    print("Bluetooth address {} is already in use - is the server already running?".format(self.server_address))
                raise e
            self._server_sock.listen(1)

            #wait for client connection
            self._conn_thread = WrapThread(target=self._wait_for_connection)
            self._conn_thread.start()

            self._running = True

    def stop(self):
        """
        Stops the Bluetooth server if its running.
        """
        if self._running:
            if self._conn_thread:
                self._conn_thread.stop()
                self._conn_thread = None

    def send(self, data):
        """
        Send data to a connected bluetooth client

        :param string data:
            The data to be sent.
        """
        if self._client_connected:
            if self._encoding:
                data = string_to_bytes(data, self._encoding)
            try:
                self._client_sock.send(data)
            except BLUETOOTH_EXCEPTIONS as e:
                self._handle_bt_error(e)

    def _wait_for_connection(self):
        #keep going until the server is stopped
        while not self._conn_thread.stopping.is_set():
            #wait for connection
            self._client_connected = False
            while not self._conn_thread.stopping.is_set() and not self._client_connected:
                try:
                    self._client_sock, self._client_info = self._server_sock.accept()
                    self._client_connected = True
                except socket.timeout as e:
                    self._handle_bt_error(e)

            #did a client connect?
            if self._client_connected:
                #call the call back
                if self.when_client_connects:
                    self.when_client_connects()
                
                #read data
                self._read()

        #server has been stopped
        self._server_sock.close()
        self._server_sock = None
        self._running = False

    def _read(self):
        #read until the server is stopped or the client disconnects
        while not self._conn_thread.stopping.is_set() and self._client_connected:
            #read data from bluetooth socket
            data = ""
            try:
                data = self._client_sock.recv(1024, socket.MSG_DONTWAIT)
            except BLUETOOTH_EXCEPTIONS as e:
                self._handle_bt_error(e)
            if len(data) > 0:
                #print("received [%s]" % data)
                if self._data_received_callback:
                    if self._encoding:
                        data = data.decode(self._encoding)
                    self.data_received_callback(data)
            sleep(BLUETOOTH_TIMEOUT)

        #close the client socket
        self._client_sock.close()
        self._client_sock = None
        self._client_info = None
        self._client_connected = False

    def _handle_bt_error(self, bt_error):
        #'timed out' is caused by the wait_for_connection loop
        if str(bt_error) == "timed out":
            return
        #'resource unavailable' is when data cannot be read because there is nothing in the buffer
        if str(bt_error) == "[Errno 11] Resource temporarily unavailable":
            return
        #'connection reset' is caused when the client disconnects
        if str(bt_error) == "[Errno 104] Connection reset by peer":
            self._client_connected = False
            if self.when_client_disconnects:
                self.when_client_disconnects()
            return
        #'conection timeout' is caused when the server can no longer connect to read from the client
        # (perhaps the client has gone out of range)
        if str(bt_error) == "[Errno 110] Connection timed out":
            self._client_connected = False
            if self.when_client_disconnects:
                self.when_client_disconnects()
            return
        raise bt_error

class BluetoothClient():
    """
    Creates a Bluetooth client which can send data to a server using RFCOMM Serial Data.

    The following example will create a bluetooth client which will connect to a paired 
    device called ``raspberrypi``, send ``helloworld`` and print any data is receives::
    
        from bluedot.btcomm import BluetoothClient
        from signal import pause
        
        def data_received(data):
            print(data)

        c = BluetoothClient("raspberrypi", data_received)
        c.send("helloworld")

        pause()
    
    :param string server:
        The server name ("raspberrypi") or server mac address 
        ("11:11:11:11:11:11") to connect too.

        The server must be a paired device.

    :param data_received_callback:
        A function reference should be passed, this function will be called when
        data is received by the client.  The function should accept a single parameter
        which when called will hold the data received. Set to ``None`` if data 
        received is not required.

    :param int port:
        The bluetooth port the client should use, the default is ``1``

    :param string device:
        The bluetooth device to be used, the default is ``hci0``, if your device 
        only has 1 bluetooth adapter this shouldn't need to be changed.

    :param string encoding:
        The encoding standard to be used when sending and receiving byte data. The default is 
        ``utf-8``.  If set to ``None`` no encoding is done and byte data types should be used.

    :param bool power_up_device:
        If ``True``, the bluetooth device will be powered up (if required) when the 
        server starts. The default is ``False``. 
        
        Depending on how bluetooth has been powered down, you may need to use rfkill 
        to unblock bluetooth to give permission to bluez to power on bluetooth::

            sudo rfkill unblock bluetooth

    :param bool auto_connect:
        If ``True`` (the default), the bluetooth client will automatically try to
        connect to the server at initialisation, if ``False``, the method ``connect`` 
        will need to be called.

    """
    def __init__(self, 
        server,
        data_received_callback,
        port = 1,
        device = "hci0", 
        encoding = "utf-8",
        power_up_device = False,
        auto_connect = True):

        self._device = device
        self._server = server
        self._data_received_callback = data_received_callback
        self._port = port
        self._power_up_device = power_up_device
        self._encoding = encoding

        self._adapter = BluetoothAdapter(self._device)

        self._connected = False
        self._client_sock = None

        self._conn_thread = None

        if auto_connect:
            self.connect()

    @property
    def device(self):
        """
        The bluetooth device the client is using. This defaults to hci0.
        """
        return self.adapter.device

    @property
    def server(self):
        """
        The server name ("raspberrypi") or server mac address 
        ("11:11:11:11:11:11") to connect too.
        """
        return self._server

    @property
    def port(self):
        """
        The port the client is using. This defaults to 1.
        """
        return self._port

    @property
    def adapter(self):
        """
        A BluetoothAdapter object which represents the bluetooth device 
        the client is using.
        """
        return self._adapter

    @property
    def encoding(self):
        """
        The encoding standard the client is using. The default is ``utf-8``. 
        """
        return self._encoding

    @property
    def client_address(self):
        """
        The mac address of the device being used.
        """
        return self.adapter.address

    @property
    def connected(self):
        """
        Returns ``True`` when connected.
        """
        return self._connected

    @property
    def data_received_callback(self):
        """
        Sets or returns the function which is called when data is received by the client. 
        
        The function should accept a single parameter which when called will hold
        the data received. Set to ``None`` if data received is not required.
        """
        return self._data_received_callback

    @data_received_callback.setter
    def data_received_callback(self, value):
        self._data_received_callback = value

    def connect(self):
        """
        Connect to a bluetooth server.
        """
        if not self._connected:

            if self._power_up_device:
                self.adapter.powered = True

            if not self.adapter.powered:
                raise Exception("Bluetooth device {} is turned off".format(self.adapter.device))

            #try and find the server name or mac address in the paired devices list
            server_mac = None
            for device in self.adapter.paired_devices:
                if self._server == device[0] or self._server == device[1]:
                    server_mac = device[0]
                    break
            if server_mac == None:
                raise Exception("Server {} not found in paired devices".format(self._server))
            
            #create a socket
            self._client_sock = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_STREAM, socket.BTPROTO_RFCOMM)
            self._client_sock.bind((self.adapter.address, self._port))
            self._client_sock.connect((server_mac, self._port))
        
            self._connected = True 
        
            self._conn_thread = WrapThread(target=self._read)
            self._conn_thread.start()

    def disconnect(self):
        """
        Disconnect from a bluetooth server.
        """
        if self._connected:
            
            #stop the connection thread
            if self._conn_thread:
                self._conn_thread.stop()
                self._conn_thread = None

            #close the socket
            try:
                self._client_sock.close()
            finally:
                self._client_sock = None
                self._connected = False
        
    def send(self, data):
        """
        Send data to a bluetooth server

        :param string data:
            The data to be sent.
        """
        if self._connected:
            if self._encoding:
                data = string_to_bytes(data, self._encoding)
            try:
                self._client_sock.send(data)
            except BLUETOOTH_EXCEPTIONS as e:
                self._handle_bt_error(e)
            
    def _read(self):
        #read until the client is stopped or the client disconnects
        while not self._conn_thread.stopping.is_set() and self._connected:
            #read data from bluetooth socket
            data = ""
            try:
                data = self._client_sock.recv(1024, socket.MSG_DONTWAIT)
            except BLUETOOTH_EXCEPTIONS as e:
                self._handle_bt_error(e)
            if len(data) > 0:
                #print("received [%s]" % data)
                if self._data_received_callback:
                    if self._encoding:
                        data = data.decode(self._encoding)
                    self.data_received_callback(data)
            sleep(BLUETOOTH_TIMEOUT)

    def _handle_bt_error(self, bt_error):
        #'resource unavailable' is when data cannot be read because there is nothing in the buffer
        if str(bt_error) == "[Errno 11] Resource temporarily unavailable":
            return
        #'connection reset' is caused when the client disconnects
        if str(bt_error) == "[Errno 104] Connection reset by peer":
            self._connected = False
            return
        #'conection timeout' is caused when the server can no longer connect to read from the client
        # (perhaps the client has gone out of range)
        if str(bt_error) == "[Errno 110] Connection timed out":
            self._connected = False
            return
        raise bt_error



import atexit
from threading import Thread, Event

_THREADS = set()

def _shutdown():
    while _THREADS:
        for t in _THREADS.copy():
            t.stop()

atexit.register(_shutdown)

class WrapThread(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        super(WrapThread, self).__init__(group, target, name, args, kwargs)
        self.stopping = Event()
        self.daemon = True

    def start(self):
        self.stopping.clear()
        _THREADS.add(self)
        super(WrapThread, self).start()

    def stop(self):
        self.stopping.set()
        self.join()

    def join(self):
        super(WrapThread, self).join()
        _THREADS.discard(self)




from setuptools import setup

__project__ = 'bluedot'
__desc__ = 'A zero boiler plate bluetooth remote'
__version__ = '1.0.4'
__author__ = "Martin O'Hanlon"
__author_email__ = 'martin@ohanlonweb.com'
__license__ = 'MIT'
__url__ = 'https://github.com/martinohanlon/BlueDot'
__requires__ = ['pydbus',]

__classifiers__ = [
#    "Development Status :: 3 - Alpha",
#    "Development Status :: 4 - Beta",
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Education",
    "Intended Audience :: Developers",
    "Topic :: Education",
    "Topic :: Communications",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
]

setup(name='bluedot',
      version = __version__,
      description = __desc__,
      url = __url__,
      author = __author__,
      author_email = __author_email__,
      license= __license__,
      packages = [__project__],
#      install_requires = __requires__,
      zip_safe=False)



from __future__ import absolute_import, print_function, unicode_literals
import dbus
import time
import sys

SERVICE_NAME = "org.bluez"
ADAPTER_INTERFACE = SERVICE_NAME + ".Adapter1"
DEVICE_INTERFACE = SERVICE_NAME + ".Device1"
PROFILE_MANAGER = SERVICE_NAME + ".ProfileManager1"

def get_managed_objects():
    bus = dbus.SystemBus()
    manager = dbus.Interface(bus.get_object(SERVICE_NAME, "/"), "org.freedesktop.DBus.ObjectManager")
    return manager.GetManagedObjects()

def find_adapter(pattern=None):
    return find_adapter_in_objects(get_managed_objects(), pattern)

def find_adapter_in_objects(objects, pattern=None):
    bus = dbus.SystemBus()
    for path, ifaces in objects.items():
        adapter = ifaces.get(ADAPTER_INTERFACE)
        if adapter is None:
            continue
        if not pattern or pattern == adapter["Address"] or path.endswith(pattern):
            obj = bus.get_object(SERVICE_NAME, path)
            return dbus.Interface(obj, ADAPTER_INTERFACE)
    raise Exception("Bluetooth adapter {} not found".format(pattern))

def get_adapter_property(device_name, property):
    bus = dbus.SystemBus()
    adapter_path = find_adapter(device_name).object_path
    adapter = dbus.Interface(bus.get_object(SERVICE_NAME, adapter_path),"org.freedesktop.DBus.Properties")
    return adapter.Get(ADAPTER_INTERFACE, property)    

def get_mac(device_name):
    return get_adapter_property(device_name, "Address")

def get_adapter_powered_status(device_name):
    powered = get_adapter_property(device_name, "Powered")
    return True if powered else False

def get_adapter_discoverable_status(device_name):
    discoverable = get_adapter_property(device_name, "Discoverable")
    return True if discoverable else False

def get_adapter_pairable_status(device_name):
    pairable = get_adapter_property(device_name, "Pairable")
    return True if pairable else False

def get_paired_devices(device_name):
    paired_devices = []

    bus = dbus.SystemBus()
    adapter_path = find_adapter(device_name).object_path
    om = dbus.Interface(bus.get_object(SERVICE_NAME, "/"), "org.freedesktop.DBus.ObjectManager")
    objects = om.GetManagedObjects()

    for path, interfaces in objects.items():
        if DEVICE_INTERFACE not in interfaces:
            continue
        properties = interfaces[DEVICE_INTERFACE]
        if properties["Adapter"] != adapter_path:
            continue

        paired_devices.append((str(properties["Address"]), str(properties["Alias"])))

    return paired_devices

def device_discoverable(device_name, discoverable):
    bus = dbus.SystemBus()
    adapter_path = find_adapter(device_name).object_path
    adapter = dbus.Interface(bus.get_object(SERVICE_NAME, adapter_path),"org.freedesktop.DBus.Properties")
    if discoverable:
        value = dbus.Boolean(1)
    else:
        value = dbus.Boolean(0)
    adapter.Set(ADAPTER_INTERFACE, "Discoverable", value)

def device_pairable(device_name, pairable):
    bus = dbus.SystemBus()
    adapter_path = find_adapter(device_name).object_path
    adapter = dbus.Interface(bus.get_object(SERVICE_NAME, adapter_path),"org.freedesktop.DBus.Properties")
    if pairable:
        value = dbus.Boolean(1)
    else:
        value = dbus.Boolean(0)
    adapter.Set(ADAPTER_INTERFACE, "Pairable", value)

def device_powered(device_name, powered):
    bus = dbus.SystemBus()
    adapter_path = find_adapter(device_name).object_path
    adapter = dbus.Interface(bus.get_object(SERVICE_NAME, adapter_path),"org.freedesktop.DBus.Properties")
    if powered:
        value = dbus.Boolean(1)
    else:
        value = dbus.Boolean(0)
    adapter.Set(ADAPTER_INTERFACE, "Powered", value)

def register_spp(port):

    service_record = """
    <?xml version="1.0" encoding="UTF-8" ?>
    <record>
      <attribute id="0x0001">
        <sequence>
          <uuid value="0x1101"/>
        </sequence>
      </attribute>

      <attribute id="0x0004">
        <sequence>
          <sequence>
            <uuid value="0x0100"/>
          </sequence>
          <sequence>
            <uuid value="0x0003"/>
            <uint8 value="{}" name="channel"/>
          </sequence>
        </sequence>
      </attribute>

      <attribute id="0x0100">
        <text value="Serial Port" name="name"/>
      </attribute>
    </record>
    """.format(port)

    bus = dbus.SystemBus()

    manager = dbus.Interface(bus.get_object(SERVICE_NAME, "/org/bluez"), PROFILE_MANAGER)

    path = "/bluez"
    uuid = "00001101-0000-1000-8000-00805f9b34fb"
    opts = {
#        "AutoConnect" : True,
        "ServiceRecord" : service_record
    }
    
    try:
        manager.RegisterProfile(path, uuid, opts)
    except dbus.exceptions.DBusException as e:
        #the spp profile has already been registered, ignore
        if str(e) != "org.bluez.Error.AlreadyExists: Already Exists":
            raise(e)

if sys.version_info[0] > 2:
    def string_to_bytes(data, encoding):
        return bytes(data, encoding=encoding)
else:
    def string_to_bytes(data, encoding):
        data.encode(encoding)
        return bytes(data)

